/**
 * Copyright 2006 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nutch.parse.xml;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.nutch.metadata.Metadata;
import org.apache.nutch.parse.Outlink;
import org.apache.nutch.parse.OutlinkExtractor;
import org.apache.nutch.parse.Parse;
import org.apache.nutch.parse.ParseData;
import org.apache.nutch.parse.ParseImpl;
import org.apache.nutch.parse.ParseStatus;
import org.apache.nutch.parse.Parser;
import org.apache.nutch.parse.xml.config.XMLField;
import org.apache.nutch.parse.xml.config.XMLIndexerProperties;
import org.apache.nutch.parse.xml.config.XMLParserConfig;
import org.apache.nutch.protocol.Content;
import org.jaxen.JaxenException;
import org.jaxen.SimpleNamespaceContext;
import org.jaxen.jdom.JDOMXPath;
import org.jdom.Attribute;
import org.jdom.CDATA;
import org.jdom.Comment;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.EntityRef;
import org.jdom.Namespace;
import org.jdom.ProcessingInstruction;
import org.jdom.Text;

/**
 * Parseur XML qui se base sur le fichier XML de configuration
 * "xmlparser-conf.xml" situ dans le rpertoire conf de Nutch pour extraire les
 * lments d'un fichier XML. Il Utilise JDOM et Jaxen pour le support XPath
 * <br/> This parser use an XML config file to extratct data from elements. The
 * config file "xmlparser-conf.xml" is in the Nutch conf directory. It use JDOM
 * and Jaxen for XPath support
 * 
 * @author Rida Benjelloun (rida.benjelloun@doculibre.com)
 */
public class XMLParser implements Parser {

	public static final Log LOG = LogFactory
			.getLog("org.apache.nutch.parse.xml");

	private SimpleNamespaceContext nsc = new SimpleNamespaceContext();

	private Configuration conf;

	private String title = "";

	public Parse getParse(Content content) {
		Metadata metadata = new Metadata();

		byte[] raw = content.getContent();
		Document doc = XMLUtils.parseXml(new ByteArrayInputStream(raw));

		String text = "";
		try {
			text = getContent(doc);

			XMLParserConfig xpc = XMLParserConfig.getInstance();

			Collection populatedFields = populateXmlFields(doc, xpc
					.getXMLIndexerProperties());
			Iterator it = populatedFields.iterator();
			while (it.hasNext()) {
				XMLField xf = (XMLField) it.next();
				metadata.set(xf.getFieldName(), "nxpToIndex_"
						+ xf.getFieldValue() + " -nxp- " + xf.getFieldType()
						+ " -nxp- " + xf.getFieldBoost());
			}

		} catch (JaxenException e) {
			return new ParseStatus(ParseStatus.FAILED,
					"XML Parser Jaxen error : " + e.getMessage())
					.getEmptyParse(getConf());
		} catch (CloneNotSupportedException e) {
			return new ParseStatus(ParseStatus.FAILED,
					"XML Parser clone() error : " + e.getMessage())
					.getEmptyParse(getConf());
		}

		Outlink[] outlinks = OutlinkExtractor.getOutlinks(text, getConf());

		ParseData parseData = new ParseData(ParseStatus.STATUS_SUCCESS, title,
				outlinks, content.getMetadata(), metadata);
		parseData.setConf(this.conf);

		return new ParseImpl(text, parseData);
	}

	public String getContent(Document doc) throws JaxenException {
		return concatNoedsContent(doc, "//*");
	}

	public String concatNoedsContent(Object xmlDoc, String xpath)
			throws JaxenException {
		StringBuffer chaineConcat = new StringBuffer();
		JDOMXPath xp = new JDOMXPath(xpath);
		xp.setNamespaceContext(nsc);
		List ls = xp.selectNodes(xmlDoc);
		Iterator i = ls.iterator();
		int j = 0;
		while (i.hasNext()) {
			j++;
			String text = "";
			Object obj = (Object) i.next();
			if (obj instanceof Element) {
				Element elem = (Element) obj;
				text = elem.getText().trim();
			} else if (obj instanceof Attribute) {
				Attribute att = (Attribute) obj;
				text = att.getValue().trim();
			} else if (obj instanceof Text) {
				Text txt = (Text) obj;
				text = txt.getText().trim();
			} else if (obj instanceof CDATA) {
				CDATA cdata = (CDATA) obj;
				text = cdata.getText().trim();
			} else if (obj instanceof Comment) {
				Comment com = (Comment) obj;
				text = com.getText().trim();
			} else if (obj instanceof ProcessingInstruction) {
				ProcessingInstruction pi = (ProcessingInstruction) obj;
				text = pi.getData().trim();
			} else if (obj instanceof EntityRef) {
				EntityRef er = (EntityRef) obj;
				text = er.toString().trim();
			}
			if (text != "") {
				if (ls.size() == 1) {
					chaineConcat.append(text);
					return chaineConcat.toString().trim();
				} else {
					if (ls.size() == j)
						chaineConcat.append(text);
					else
						chaineConcat.append(text + " ");
				}
			}
		}

		return chaineConcat.toString().trim();
	}

	public Collection populateXmlFields(Object xml, Map xmlIndexersProperties)
			throws JaxenException, CloneNotSupportedException {
		List documentNs = null;
		boolean nsProcessed = false;
		Collection populatedFields = new ArrayList();
		documentNs = getAllDocumentNs((org.jdom.Document) xml);
		if (documentNs.size() > 0) {
			for (int j = 0; j < documentNs.size(); j++) {
				Collection xmlFields = null;
				if (xmlIndexersProperties.containsKey(documentNs.get(j))) {
					nsProcessed = true;
					xmlFields = ((XMLIndexerProperties) xmlIndexersProperties
							.get(documentNs.get(j))).getXmlFields();
					extractDataFromElements(xml, xmlFields, populatedFields);
				}
			}
		}

		if (nsProcessed == false) {
			if (xmlIndexersProperties.containsKey("default")) {
				Collection xmlConcatfields = ((XMLIndexerProperties) xmlIndexersProperties
						.get("default")).getXmlFields();
				Iterator it = xmlConcatfields.iterator();
				while (it.hasNext()) {
					XMLField cf = (XMLField) it.next();
					XMLField ncf = (XMLField) cf.clone();
					ncf.setFieldValue(concatNoedsContent(xml, cf
							.getFieldXPath()));
					populatedFields.add(ncf);
				}
			}
		}

		return populatedFields;
	}

	private void extractDataFromElements(Object xmlDoc, Collection xmlFields,
			Collection resColl) throws JaxenException,
			CloneNotSupportedException {
		Iterator it = xmlFields.iterator();
		while (it.hasNext()) {
			Object field = it.next();
			if (field instanceof XMLField) {
				XMLField lf = (XMLField) field;
				JDOMXPath xp = new JDOMXPath(lf.getFieldXPath());
				xp.setNamespaceContext(nsc);
				List selectList = xp.selectNodes(xmlDoc);
				StringBuffer elementValues = new StringBuffer();
				Iterator i = selectList.iterator();
				XMLField lfoccur = (XMLField) lf.clone();
				while (i.hasNext()) {					
					Object selection = (Object) i.next();
					if (selection instanceof Element) {
						Element elem = (Element) selection;
						if (elem.getText().trim() != null
								&& elem.getText().trim() != "") {
							elementValues.append(elem.getText()+" ");							
						}
					} else if (selection instanceof Attribute) {
						Attribute att = (Attribute) selection;
						elementValues.append(att.getValue()+" ");						
					} else if (selection instanceof Text) {
						Text text = (Text) selection;
						elementValues.append(text.getText()+" ");						
					} else if (selection instanceof CDATA) {
						CDATA cdata = (CDATA) selection;
						elementValues.append(cdata.getText()+" ");						
					} else if (selection instanceof Comment) {
						Comment com = (Comment) selection;
						elementValues.append(com.getText()+" ");						
					} else if (selection instanceof ProcessingInstruction) {
						ProcessingInstruction pi = (ProcessingInstruction) selection;
						elementValues.append(pi.getData()+" ");						
					} else if (selection instanceof EntityRef) {
						EntityRef er = (EntityRef) selection;
						elementValues.append(er.toString()+" ");
						resColl.add(lfoccur);
					}
					if (lfoccur.getFieldName().equals("title")
							&& lfoccur.getFieldValue() != null) {
						title = lfoccur.getFieldValue();
					}
				}
				lfoccur.setFieldValue(elementValues.toString());
				resColl.add(lfoccur);

			} else {
				resColl.add(field);
			}
		}

	}

	// private void extractDataFromElements(Object xmlDoc, Collection xmlFields,
	// Collection resColl) throws JaxenException,
	// CloneNotSupportedException {
	// Iterator it = xmlFields.iterator();
	// while (it.hasNext()) {
	// Object field = it.next();
	// if (field instanceof XMLField) {
	// XMLField lf = (XMLField) field;
	// JDOMXPath xp = new JDOMXPath(lf.getFieldXPath());
	// xp.setNamespaceContext(nsc);
	// List selectList = xp.selectNodes(xmlDoc);
	// Iterator i = selectList.iterator();
	// while (i.hasNext()) {
	// XMLField lfoccur = (XMLField) lf.clone();
	// Object selection = (Object) i.next();
	// if (selection instanceof Element) {
	// Element elem = (Element) selection;
	// if (elem.getText().trim() != null
	// && elem.getText().trim() != "") {
	// lfoccur.setFieldValue(elem.getText());
	// resColl.add(lfoccur);
	// }
	// } else if (selection instanceof Attribute) {
	// Attribute att = (Attribute) selection;
	// lfoccur.setFieldValue(att.getValue());
	// resColl.add(lfoccur);
	// } else if (selection instanceof Text) {
	// Text text = (Text) selection;
	// lfoccur.setFieldValue(text.getText());
	// resColl.add(lfoccur);
	// } else if (selection instanceof CDATA) {
	// CDATA cdata = (CDATA) selection;
	// lfoccur.setFieldValue(cdata.getText());
	// resColl.add(lfoccur);
	// } else if (selection instanceof Comment) {
	// Comment com = (Comment) selection;
	// lfoccur.setFieldValue(com.getText());
	// resColl.add(lfoccur);
	// } else if (selection instanceof ProcessingInstruction) {
	// ProcessingInstruction pi = (ProcessingInstruction) selection;
	// lfoccur.setFieldValue(pi.getData());
	// resColl.add(lfoccur);
	// } else if (selection instanceof EntityRef) {
	// EntityRef er = (EntityRef) selection;
	// lfoccur.setFieldValue(er.toString());
	// resColl.add(lfoccur);
	// }
	// if (lfoccur.getFieldName().equals("title")
	// && lfoccur.getFieldValue() != null) {
	// title = lfoccur.getFieldValue();
	// }
	// }
	//
	// } else {
	// resColl.add(field);
	// }
	// }
	//
	// }
	//

	public List getAllDocumentNs(org.jdom.Document doc) {
		List ls = new ArrayList();
		processChildren(doc.getRootElement(), ls);
		return ls;
	}

	private boolean exist(List nsLs, String nsUri) {
		if (nsLs.isEmpty())
			return false;
		for (int i = 0; i < nsLs.size(); i++) {
			if (((String) nsLs.get(i)).equals(nsUri)) {
				return true;
			}
		}
		return false;
	}

	private void processChildren(Element elem, List ns) {
		Namespace nsCourent = (Namespace) elem.getNamespace();
		String nsUri = (nsCourent.getURI());
		if (!exist(ns, nsUri)) {
			ns.add(nsUri.trim());
			nsc.addNamespace(nsCourent.getPrefix(), nsCourent.getURI());
		}
		List additionalNs = elem.getAdditionalNamespaces();
		if (!additionalNs.isEmpty())
			copyNsList(additionalNs, ns);
		if (elem.getChildren().size() > 0) {
			List elemChildren = elem.getChildren();
			for (int i = 0; i < elemChildren.size(); i++) {
				processChildren((Element) elemChildren.get(i), ns);
			}
		}
	}

	private void copyNsList(List nsElem, List nsRes) {
		for (int i = 0; i < nsElem.size(); i++) {
			Namespace ns = (Namespace) nsElem.get(i);
			nsc.addNamespace(ns.getPrefix(), ns.getURI());
			nsRes.add(ns.getURI().trim());
		}
	}

	public void setConf(Configuration conf) {
		this.conf = conf;
	}

	public Configuration getConf() {
		return this.conf;
	}

}
